// Copyright 2011-2012 Daniele Panozzo, panozzo@inf.ethz.ch. All rights reserved.
// The code in this file cannot be used for commercial
// applications without prior written permission from Daniele Panozzo

//#ifdef TIMER
#include "timer.h"
//#endif

#include <GL/glew.h>
#include "state.h"
#include <QtOpenGL/QGLWidget>

#include "solver25x25/solver_25x25_interface.h"

#undef min
#undef max
#include <limits>
#include <math.h>

#include <fstream>
#include <iostream>

extern void display ();

// Global static pointer used to ensure a single instance of the class.
State* State::m_pInstance = NULL; 

/** This function is called to create an instance of the class.
 Calling the constructor publicly is not allowed. The constructor
 is private and is only called by this Instance function.
 */

State* State::Instance()
{
    if (!m_pInstance)   // Only allow one instance of class to be generated.
        m_pInstance = new State;
    return m_pInstance;
}

// compute the next highest power of 2 of 32-bit v
unsigned int State::roundPow2(unsigned int v)
{
	v--;
	v |= v >> 1;
	v |= v >> 2;
	v |= v >> 4;
	v |= v >> 8;
	v |= v >> 16;
	v++;
	return v;
}

void State::loadImage(char* imgfile)
{
	// the image is loaded in a matrix whose size is a power of two
	// the image starts at 0,0 and and ends at its original size WHEN used as an opengl texture
	
    qImage.load(QString(imgfile));
	
    filename = string(imgfile);
    
    // Init Grid
    
    originalSize[0] = currentSize[0] = qImage.width();
    originalSize[1] = currentSize[1] = qImage.height();
    
    // Init Crop
    cropBL(0) = 0;
    cropBL(1) = 0;
    
    cropTR(0) = originalSize[0];
    cropTR(1) = originalSize[1];

    crop2P[0] = 0;
    crop2P[1] = 0;
    
    initGrid();
    
	tex = QGLWidget::convertToGLFormat( qImage );
	
	// allocate a texture name
    
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    
	glGenTextures( 1, &textureImage );
	glGenTextures( 1, &textureImageLinear );
	
	// select our current texture
	glBindTexture( GL_TEXTURE_2D, textureImage );
	
    glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    glTexImage2D(GL_TEXTURE_2D, 0, 3, tex.width(), tex.height(), 0, GL_RGBA,
				 GL_UNSIGNED_BYTE, tex.bits());

	glBindTexture( GL_TEXTURE_2D, textureImageLinear );
    
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	
	glTexImage2D(GL_TEXTURE_2D, 0, 3, tex.width(), tex.height(), 0, GL_RGBA,
				 GL_UNSIGNED_BYTE, tex.bits());
	
}

void State::loadSaliency(char* imgfile)
{
	
    QImage tmp;
    tmp.load(QString(imgfile));
    cerr << "Loading saliency: " << imgfile << endl;
	tmp = tmp.scaled(originalSize[0], originalSize[1]);
    
    for(int i=0; i<originalSize[0]; ++i)
    {
        for(int j=0; j<originalSize[1];++j)
        {
            QRgb rgbt = tmp.pixel(i,j);
            QColor rgb(rgbt);
            int val = (rgb.red() + rgb.green() + rgb.blue()) / 3;
            setSalLoadedOriginal(i, (originalSize[1]-1) - j, val);
        }
    }
    
    updateLoaded();
    updateSaliency();
}

State::State()
{
    XCELLS = 25;
    YCELLS = 25;
    
    brushRadius = 120;
    brushIntensity = 256;
    
    LWfactor = 0.15;
    LHfactor = 0.15;
    
    energyType = ENERGY_ASAP;
    saliencyType = SAL_BOTH;
    
    autoSolve = true;
    
    laplacianWeight = 0.5;
    addLaplacianTermsEnabled = false;
    
    HYBRID_alpha = 0.96;
    
    useExtGrid = true;
    resExtGrid = 2;
    
    eraserActive = false;
    brushSmooth = true;
    
    exportMode = false;
    
    brightness = 0;
    contrast = 0;
}

void State::init()
{


    initGrid();
    initSaliency();    

}


void State::initGrid()
{
    // the grid is XCELLSxYCELLS quad so we have XCELLS+1 x XCELLS+1 vertices
    int rows = XCELLS+1;
    int cols = YCELLS+1;
    
    gridX = VectorXd::Zero(rows);
    gridY = VectorXd::Zero(cols);
    gridU = VectorXd::Zero(rows);
    gridV = VectorXd::Zero(cols);

    for(int i=0; i<rows; ++i)
    {
        gridX(i) = i * (originalSize[0]/(rows-1));
        gridU(i) = (double)i / double(rows-1);
    }   

    for (int j=0; j<cols; ++j) 
    {
        gridY(j) = j * (originalSize[1]/(cols-1));
        gridV(j) = (double)j / double(cols-1);
    }
    
    gridXOri = gridX;
    gridYOri = gridY;
}

void State::updateSaliencyGrid()
{
    // the grid is XCELLSxYCELLS quad so we have XCELLS+1 x XCELLS+1 vertices
    int rows = XCELLS+1;
    int cols = YCELLS+1;

//    for(int i=0; i<rows; ++i)
//        saliencyX(i) = i * (originalSize[0]/(rows-1)); // TODO: CHECK HERE
//    
//    for (int j=0; j<cols; ++j) 
//        saliencyY(j) = j * (originalSize[1]/(cols-1));

    Vector2d cropSize = cropTR - cropBL;
    
    for(int i=0; i<rows; ++i)
        saliencyX(i) = cropBL(0) + i * (cropSize[0]/(rows-1));
    
    for (int j=0; j<cols; ++j) 
        saliencyY(j) = cropBL(1) + j * (cropSize[1]/(cols-1));
    
}

void State::initSaliency(bool clear)
{
    if (clear)
    {
        saliency.resize(originalSize[0] * originalSize[1]);
        saliencyLoaded.resize(originalSize[0] * originalSize[1]);
        saliencyLoadedOriginal.resize(originalSize[0] * originalSize[1]);
        for(int i=0; i<originalSize[0] * originalSize[1]; ++i)
        {
            saliency[i] = 1;
            saliencyLoaded[i] = 1;
            saliencyLoadedOriginal[i] = 1;
        }
    }

    // the grid is XCELLSxYCELLS quad so we have XCELLS+1 x XCELLS+1 vertices
    int rows = XCELLS+1;
    int cols = YCELLS+1;
    
    saliencyX = VectorXi::Zero(rows);
    saliencyY = VectorXi::Zero(cols);

    updateSaliencyGrid();
    
    glGenTextures( 1, &textureSaliency);
    glGenTextures( 1, &textureSaliencyLoaded);
    
    integratedSaliency = MatrixXd(rows-1,cols-1);
    integratedSaliencyValid = MatrixXd::Zero(rows-1,cols-1);
    
	updateSaliencyLocal();
}

void State::updateLoaded()
{
    double mean = 0;
    for(int i=0; i<originalSize[0]; ++i)
        for(int j=0; j<originalSize[1];++j)
            mean += ((double)getSalLoadedOriginal(i,j)/255.0);
    mean /= originalSize[0] * originalSize[1];

    for(int i=0; i<originalSize[0]; ++i)
    {
        for(int j=0; j<originalSize[1];++j)
        {
            // cout << "---: " << getSalLoadedOriginal(i,j)/255.0;
            
            double value = (double)getSalLoadedOriginal(i,j)/255.0;

            // apply contrast and brightness
            value = (value - mean) * (contrast+1.0) + mean + brightness;
            value = value < 0 ? 0 : value;
            value = value > 1 ? 1 : value;
            // cout << "--> " << value << endl;
            setSalLoaded(i, j, value*254.0);
        }
    }
}


void State::updateSaliency()
{
	// select our current texture
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glBindTexture( GL_TEXTURE_2D, textureSaliency );
	
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	
	glTexImage2D(GL_TEXTURE_2D, 0, 1, originalSize[0], originalSize[1], 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, &saliency[0]);
    
    glBindTexture( GL_TEXTURE_2D, textureSaliencyLoaded );
	
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	
	glTexImage2D(GL_TEXTURE_2D, 0, 1, originalSize[0], originalSize[1], 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, &saliencyLoaded[0]);

    
    // integrate saliency
    for(int i=0; i<saliencyX.size()-1;++i)
    {
        for(int j=0; j<saliencyY.size()-1; ++j)
        {
            double acc = 0;
            double count = 0;
            for(int x=saliencyX[i]; x < saliencyX[i+1]; ++x)
            {
                for(int y=saliencyY[j]; y < saliencyY[j+1]; ++y)
                {
                    switch(saliencyType)
                    {
                        case SAL_USER:
                            acc += getSal(x,y);
                            break;
                        case SAL_LOADED:
                            acc += getSalLoaded(x,y);
                            break;
                        case SAL_BOTH:
                            acc += getSalLoaded(x,y);
                            acc += getSal(x,y);
                            break;
                    }
                    count = count+1;
                }
            }
            integratedSaliency(i,j) = sqrt(acc/count);
            integratedSaliencyValid(i,j) = 1.0;
        }
    }
    
}

void State::updateSaliencyLocal()
{
	// select our current texture
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    
	glBindTexture( GL_TEXTURE_2D, textureSaliency );
	
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	
	glTexImage2D(GL_TEXTURE_2D, 0, 1, originalSize[0], originalSize[1], 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, &saliency[0]);

    glBindTexture( GL_TEXTURE_2D, textureSaliencyLoaded );
	
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	
	glTexImage2D(GL_TEXTURE_2D, 0, 1, originalSize[0], originalSize[1], 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, &saliencyLoaded[0]);

    // integrate saliency
    for(int i=0; i<saliencyX.size()-1;++i)
    {
        for(int j=0; j<saliencyY.size()-1; ++j)
        {
            if (integratedSaliencyValid(i,j) == 0)
            {
                double acc = 0;
                double count = 0;
                for(int x=saliencyX[i]; x < saliencyX[i+1]; ++x)
                {
                    for(int y=saliencyY[j]; y < saliencyY[j+1]; ++y)
                    {
                        switch(saliencyType)
                        {
                            case SAL_USER:
                                acc += getSal(x,y);
                                break;
                            case SAL_LOADED:
                                acc += getSalLoaded(x,y);
                                break;
                            case SAL_BOTH:
                                acc += getSalLoaded(x,y);
                                acc += getSal(x,y);
                                break;
                        }
                        count = count+1;
                    }
                }
                integratedSaliency(i,j) = sqrt(acc/count);
                integratedSaliencyValid(i,j) = 1.0;
            }
        }
    }
    
}

void State::updateCrop()
{
    updateSaliencyGrid();
    //    updateLoaded();
    updateSaliency();
}

void State::moveCrop()
{
    Vector2d& p = State::Instance()->crop2P;
    Vector2d& m = State::Instance()->cropBL;
    Vector2d& M = State::Instance()->cropTR;
    
    Vector2d& oriSize = State::Instance()->originalSize;
    Vector2d& curSize = State::Instance()->currentSize;
    
    double oriRatio = oriSize[0]/oriSize[1];
    double curRatio = curSize[0]/curSize[1];
    
    if (oriRatio>curRatio)
    {
        // resize over W
        m[1] = 0;
        M[1] = oriSize[1];
        m[0] = p[0]*oriSize[0];
        M[0] = oriSize[0] - p[1]*oriSize[0];
    }
    else
    {
        // resize over H
        m[0] = 0;
        M[0] = oriSize[0];
        m[1] = p[0]*oriSize[1];
        M[1] = oriSize[1] - p[1]*oriSize[1];
    }
    
}

void State::clearSaliency()
{
    for(int i=0; i<originalSize[0] * originalSize[1]; ++i)
        saliency[i] = 1;
    updateSaliency();
    solve();
}

void State::setSal(int x, int y, unsigned char val)
{
    saliency[(originalSize[0] * y) + x] = val;
    // mark the corresponding saliency cell as invalid
    int rows = XCELLS+1;
    int cols = YCELLS+1;
    
    if (x >= cropBL[0] && x < cropTR[0] && 
        y >= cropBL[1] && y < cropTR[1])
    {
        Vector2d tempsize = cropTR - cropBL;
        double xStep = tempsize[0]/(rows-1);
        double yStep = tempsize[1]/(cols-1);
        int i = floor((x-cropBL[0])/xStep);
        int j = floor((y-cropBL[1])/yStep);
        integratedSaliencyValid(i,j) = 0.0;
    }
}

void State::setSalLoaded(int x, int y, unsigned char val)
{
    saliencyLoaded[(originalSize[0] * y) + x] = val;
}

void State::setSalLoadedOriginal(int x, int y, unsigned char val)
{
    saliencyLoadedOriginal[(originalSize[0] * y) + x] = val;
}

double State::getSal(int x, int y)
{
    double value = saliency[(originalSize[0] * y) + x];
    
//    // apply contrast and brightness
//    if (brightness < 0.0)  value = value * ( 1.0 + brightness);
//    else value = value + ((1 - value) * brightness);
//    value = (value - 0.5) * (tan ((contrast + 1) * M_PI/4) ) + 0.5;
    
    return value;
}

double State::getSalLoaded(int x, int y)
{
    double value = saliencyLoaded[(originalSize[0] * y) + x];
    
//    // apply contrast and brightness
//    if (brightness < 0.0)  value = value * ( 1.0 + brightness);
//    else value = value + ((1 - value) * brightness);
//    value = (value - 0.5) * (tan ((contrast + 1) * M_PI/4) ) + 0.5;

    return value;
}

double State::getSalLoadedOriginal(int x, int y)
{
    double value = saliencyLoadedOriginal[(originalSize[0] * y) + x];
    
    //    // apply contrast and brightness
    //    if (brightness < 0.0)  value = value * ( 1.0 + brightness);
    //    else value = value + ((1 - value) * brightness);
    //    value = (value - 0.5) * (tan ((contrast + 1) * M_PI/4) ) + 0.5;
    
    return value;
}

void State::solve(bool force, bool render)
{
    // fix invalid value for the minimal limit
    
    double wori = originalSize[0];
    double w = currentSize[0];
    double maxLW = (w/wori) - 0.01;
    if (LWfactor > maxLW)
        LWfactor = maxLW;
    
    double hori = originalSize[1];
    double h = currentSize[1];
    double maxLH = (h/hori) - 0.01;
    if (LHfactor > maxLH)
        LHfactor = maxLH;

#ifdef TIMER
    Timer t,t2;
    t2.start();
    t.start();
#endif
    
    if (force || autoSolve)
    {
        
        State* st = State::Instance();
        
        vector<double> S((XCELLS+YCELLS)*(XCELLS+YCELLS),1);
        double W = st->currentSize[0];
        double H = st->currentSize[1];
        double WOri = st->originalSize[0];
        double HOri = st->originalSize[1];
        double LW = LHfactor*(double)WOri/XCELLS;
        double LH = LWfactor*(double)HOri/YCELLS;
        
        MatrixXd A;
        VectorXd b;
        
        switch(energyType)
        {
            case ENERGY_ASAP:
                generateASAPEnergySystem(A, b, addLaplacianTermsEnabled);
                break;            
            case ENERGY_ARAP:
                generateARAPEnergySystem(A, b, addLaplacianTermsEnabled);
                break;
            case ENERGY_HYBRID:
                generateHYBRIDEnergySystem(A, b, HYBRID_alpha, addLaplacianTermsEnabled);
                break;            
            default:
                generateASAPEnergySystem(A, b, addLaplacianTermsEnabled);
                break;
        }
        
        MatrixXd M = A.transpose() * A;
        // Copy M in S in column-major order
        for(int j=0; j<M.cols();++j)
            for(int i=0; i<M.rows();++i)
                S[j*M.cols() + i] = M(i,j);
        
        VectorXd BTemp = -2.0*b.transpose()*A; 
        vector<double> B((XCELLS+YCELLS),0);
        assert(BTemp.rows() == (XCELLS+YCELLS));
        for(int i=0; i<BTemp.rows();++i)
            B[i] = BTemp(i);
        
        vector<double> x = Solver25x25Interface::solve(S,B,W,H,LW,LH);
        
        // the first XCELLS elements corresponds to x differences
        st->gridX[0] = 0;
        for(int i=0; i<XCELLS;++i)
        {
            st->gridX[1+i] = st->gridX[i]+x[i];
        }
        
        // the next YCELLS elements corresponds to y differences
        st->gridY[0] = 0;
        for(int i=0; i<YCELLS;++i)
            st->gridY[1+i] = st->gridY[i]+x[i+XCELLS];
    }
    
#ifdef TIMER
    double x = t.get_time();
    t.stop();
#endif
    generateInterpolatedGrid(resExtGrid);

#ifdef TIMER
    double x2 = t2.get_time();
    t2.stop();
    cerr << "Solve time: " << x << endl;
    cerr << "Solve with spline: " << x2 << endl;
#endif
    
    if (render)
        display();
}

void State::solveUniform()
{
    // the grid is XCELLSxYCELLS quad so we have XCELLS+1 x XCELLS+1 vertices
    int rows = XCELLS+1;
    int cols = YCELLS+1;
    
    gridX = VectorXd::Zero(rows);
    gridY = VectorXd::Zero(cols);
    
    for(int i=0; i<rows; ++i)
    {
        gridX(i) = i * (currentSize[0]/(rows-1));
    }   
    
    for (int j=0; j<cols; ++j) 
    {
        gridY(j) = j * (currentSize[1]/(cols-1));
    }
    
    generateInterpolatedGrid(resExtGrid);
	
	display();
}

void State::addLaplacianTerms(MatrixXd& A, VectorXd& b)
{
    int newcountX = XCELLS-1;
    int newcountY = YCELLS-1;
    
    int newcount = newcountX + newcountY;
    
    long startX = A.rows();
    long startY = A.rows() + newcountX;
    
    // Extend A
    MatrixXd Aext = MatrixXd::Zero(A.rows()+newcount,A.cols());
    Aext.block(0,0,A.rows(),A.cols()) = A;
    A = Aext;
    
    // Extend B
    VectorXd bext = VectorXd::Zero(b.size()+newcount);
    bext.head(b.size()) = b;
    b = bext;
    
    // Add new terms (b is 0 for the new terms!)
    for(int i=0; i<newcountX; ++i)
    {
        A(i+startX, i) = laplacianWeight;
        A(i+startX, i+1) = -laplacianWeight;
    }
    for(int i=0; i<newcountY; ++i)
    {
        A(i+startY, i+XCELLS) = laplacianWeight;
        A(i+startY, i+XCELLS+1) = -laplacianWeight;
    }
}

void State::generateARAPEnergySystem(MatrixXd& A, VectorXd& b, bool addLapTerms)
{
    State* st = State::Instance();
    
    Vector2d cropSize = cropTR-cropBL;
    
    double WOri = cropSize[0];
    double HOri = cropSize[1];
    
    double s0 = 1.0/((double)WOri/XCELLS);
	double t0 = 1.0/((double)HOri/YCELLS);
    
    A = MatrixXd::Zero((XCELLS*YCELLS)*2,(XCELLS+YCELLS));
    b = VectorXd::Zero((XCELLS*YCELLS)*2);
    
    for(int i=0; i < XCELLS; ++i)
    {
        for(int j=0; j < YCELLS; ++j)
        {
            int row1 = 2 * (i*XCELLS+j);
            int row2 = row1 + 1;
            double intSal = st->integratedSaliency(i,j)/255.0;
            A(row1,i)        = s0   * intSal;
            b(row1)          = 1.0  * intSal;
            A(row2,j+XCELLS) = t0   * intSal;
            b(row2)          = 1.0  * intSal;
        }
    }
    
    if (addLapTerms)
    {
        addLaplacianTerms(A, b);
    }
}

void State::generateASAPEnergySystem(MatrixXd& A, VectorXd& b, bool addLapTerms)
{
    State* st = State::Instance();
    
    Vector2d cropSize = cropTR-cropBL;

    double WOri = cropSize[0];
    double HOri = cropSize[1];
    
    double s0 = 1.0/((double)WOri/XCELLS);
	double t0 = 1.0/((double)HOri/YCELLS);
    
    A = MatrixXd::Zero((XCELLS*YCELLS),(XCELLS+YCELLS));
    b = VectorXd::Zero((XCELLS*YCELLS));

    for(int i=0; i < XCELLS; ++i)
    {
        for(int j=0; j < YCELLS; ++j)
        {
            int row = i*XCELLS+j;
            A(row,i)        = s0 *  st->integratedSaliency(i,j);
            A(row,j+XCELLS) = t0 * -st->integratedSaliency(i,j);
        }
    }
    
    if (addLapTerms)
    {
        addLaplacianTerms(A,b);
    }

}

void State::generateHYBRIDEnergySystem(MatrixXd& A, VectorXd& b, double alpha, bool addLapTerms)
{
    MatrixXd ASAP_A, ARAP_A;
    VectorXd ASAP_b, ARAP_b;
    
    generateASAPEnergySystem(ASAP_A, ASAP_b, false);
    generateARAPEnergySystem(ARAP_A, ARAP_b, false);

    int indexASAP = 0;
    long indexARAP = ASAP_A.rows();

    A = MatrixXd::Zero(ASAP_A.rows()+ARAP_A.rows(), ASAP_A.cols());
    A.block(indexASAP,0,ASAP_A.rows(),ASAP_A.cols()) = ASAP_A * (1.0-alpha);
    A.block(indexARAP,0,ARAP_A.rows(),ARAP_A.cols()) = ARAP_A * (alpha);
    
    b = VectorXd::Zero(ASAP_b.rows()+ARAP_b.rows());
    b.segment(indexASAP,ASAP_b.size()) = ASAP_b * (1.0-alpha);
    b.segment(indexARAP,ARAP_b.size()) = ARAP_b * alpha;
    
    if (addLapTerms)
        addLaplacianTerms(A, b);
    
    
}

double State::CubicInterpolate(double y0,double y1,double y2,double y3,double mu)
{
    double a0,a1,a2,a3,mu2;
    
    mu2 = mu*mu;
    a0 = y3 - y2 - y0 + y1;
    a1 = y0 - y1 - a0;
    a2 = y2 - y0;
    a3 = y1;
    
    return(a0*mu*mu2+a1*mu2+a2*mu+a3);
}

double State::SplineCubicInterpolate(double y0,double y1,double y2,double y3,double t)
{
    // TODO: CACHE ALL OF THIS
    VectorXd ts(4);
    ts << t*t*t, t*t, t, 1;
    
    MatrixXd M(4,4);
    M << -1,  3, -3,  1,
          3, -6,  3,  0,
         -3,  0,  3,  0,
          1,  4,  1,  0;
    
    VectorXd ps(4);
    ps << y0, y1, y2, y3;
    
    MatrixXd res = (ts.transpose()*(1.0/6.0))*(M*ps);
    return res(0);
}

double State::LERP(double y0, double y1, double mu)
{
    return mu*y1 + (1-mu)*y0;
}

void State::generateInterpolatedGrid_aux(VectorXd& gridX, VectorXd& gridU, VectorXd& extGridX, VectorXd& extGridU, double nsplit)
{
    extGridX = VectorXd::Zero((gridX.size()-1)*nsplit + 1);
    extGridU = VectorXd::Zero((gridU.size()-1)*nsplit + 1);
    int current = 0;
    
    for(int i=0; i<gridX.size()-1; ++i)
    {
        int is[4] = {i-1, i, i+1, i+2};
        
        // prepare coords
        double coords[4];
        for(int j=0;j<4;++j)
        {
            if (is[j] < 0)
            {
                coords[j] = -gridX[1];
            }
            else if (is[j] >= gridX.size())
            {
                long last = gridX.size()-1;
                coords[j] = gridX[last] + (gridX[last]-gridX[last-1]);
            }
            else
            {
                coords[j] = gridX[is[j]];
            }
        }
        
        // prepare uvs
        double us[2] = {gridU[i],gridU[i+1]};
        
        // finally produce the new control points
        for(int j=0; j<nsplit; ++j)
        {
            double p = (1.0/nsplit)*j;
            extGridX[current] = SplineCubicInterpolate(coords[0], coords[1], coords[2], coords[3], p);
            extGridU[current] = LERP(us[0], us[1], p);
            ++current;
        }
        
    }
    
    // add the last point
    extGridX[current] = gridX[gridX.size()-1];
    extGridU[current] = gridU[gridU.size()-1];
    
//    cerr << gridX.transpose() << endl;
//    cerr << extGridX.transpose() << endl;
//    cerr << "---------" << endl;
//    cerr << gridU.transpose() << endl;
//    cerr << extGridU.transpose() << endl;

}

void State::generateInterpolatedGrid(int n)
{
    double nsplit = std::pow(2.0,n);
    generateInterpolatedGrid_aux(gridX, gridU, extGridX, extGridU, nsplit);
    generateInterpolatedGrid_aux(gridY, gridV, extGridY, extGridV, nsplit);
}

double c(QRgb q)
{
    int r = qRed(q);
    int g = qGreen(q);
    int b = qBlue(q);
    return (r + g + b)/(255.0 * 3);
}

void State::computeGradient(const char* str)
{
    QImage q;
    q.load(QString(str));

    int w = q.width();
    int h = q.height();
    MatrixXd xGrad = MatrixXd::Zero(w,h);
    MatrixXd yGrad = MatrixXd::Zero(w,h);
    
    MatrixXd Gx(3,3);
    Gx << -1, -2, -1,
           0,  0,  0,
           1,  2,  2;
    
    MatrixXd Gy(3,3);
    Gy << -1,  0,  1,
          -2,  0,  2,
          -1,  0,  1;
    
    for(int i=0; i<w;++i)
    {
        for(int j=0;j<h;++j)
        {
            int xi[3] = {i-1,i,i+1};
            int yi[3] = {j-1,j,j+1};
            
            for(int t=0;t<3;++t)
            {
                xi[t] = xi[t] < 0 ? 0 : xi[t];
                yi[t] = yi[t] < 0 ? 0 : yi[t];
                xi[t] = xi[t] >= w ? w-1 : xi[t];
                yi[t] = yi[t] >= h ? h-1 : yi[t];
            }
            
            MatrixXd A(3,3);
            for(int u=0;u<3;++u)
                for(int v=0;v<3;++v)
                    A(u,v) = c(q.pixel(xi[u],yi[v]));
            
            xGrad(i,j) = (A.array() * Gx.array()).sum();
            yGrad(i,j) = (A.array() * Gy.array()).sum();
        }
    }
    
    xGrad = xGrad/xGrad.maxCoeff();
    yGrad = yGrad/yGrad.maxCoeff();

    for(int i=0; i<w;++i)
    {
        for(int j=0;j<h;++j)
        {
            //            cerr << i << " " << j << " " << sqrt(xGrad(i,j)*xGrad(i,j) + yGrad(i,j)*yGrad(i,j)) << endl;
            double temp = sqrt(xGrad(i,j)*xGrad(i,j) + yGrad(i,j)*yGrad(i,j))*250;
            unsigned char v = floor(temp);
            setSal(i, h-j-1, v);
        }
    }
    
    for(int i=0; i<w;++i)
    {
        for(int j=0;j<h;++j)
        {
            if (getSal(i,j) < 100)
                setSal(i,j,1);
        }
    }
    updateSaliency();
}

void State::saveSaliency(const char* str)
{
    double w = originalSize(0);
    double h = originalSize(1);
    
    QImage image(w, h, QImage::Format_RGB32);
    
    for (int x=0; x < w; ++x)
    {
        for (int y=0; y < h; ++y)
        {
            double temp = getSalLoaded(x,y) + getSal(x,y);
            if (temp > 254)
                temp = 254;
            
            image.setPixel(x, h-y-1, qRgb(temp,temp,temp));
        }
    }
    
    image.save(QString(str));
}

void State::saveParams(const char* str)
{
	string energyName;

	switch(energyType)
	{
	case ENERGY_ARAP:
		energyName = "As Rigid As Possible";
		break;

	case ENERGY_ASAP:
		energyName = "As Similar As Possible";
		break;

	case ENERGY_HYBRID:
		energyName = "Hybrid - ASAP+ARAP";
		break;
	}

	
    ofstream o(str);
    o.precision(5);
    o << filename << endl;
    o << "Original size: " << originalSize[0] << " x " << originalSize[1] << endl;
    o << "New size: " << currentSize[0] << " x " << currentSize[1] << endl;
    o << "Energy Type: " << energyName << endl;

	if(energyType == ENERGY_HYBRID)
	{
		o << "Hybrid Weight: " << HYBRID_alpha << endl;
	}

	if(addLaplacianTermsEnabled)
	{
		o << "Laplacian Weight: " << laplacianWeight << endl;
	}
	else
	{
		o << "Laplacian term disabled" << endl;
	}

    o << "Minimal quad sizes: " << LWfactor << " | " << LHfactor << endl;
	o << "Crop: " << crop2P[0] << " | " << crop2P[1] << endl;
	o << "Grid size: " << XCELLS << " x " << YCELLS << endl;
    o << "B-Spline smoothing: " << (useExtGrid ? "Enabled" : "Disabled") << endl;
    o << "Brightness(loaded saliency): " << brightness << endl;
    o << "Contrast  (loaded saliency): " << contrast << endl;
    o.close();
    
}

void State::saveMap(const char* str)
{
    ofstream o(str);
    o.precision(10);
    
    for(int i=0; i<gridXOri.size(); ++i)
        o << gridXOri[i] << " ";
    o << endl;
    
    for(int i=0; i<gridYOri.size(); ++i)
        o << gridYOri[i] << " ";
    o << endl;
    
    for(int i=0; i<gridX.size(); ++i)
        o << gridX[i] << " ";
    o << endl;
    
    for(int i=0; i<gridY.size(); ++i)
        o << gridY[i] << " ";
    o << endl;

    for(int i=0; i<extGridX.size(); ++i)
        o << extGridX[i] << " ";
    o << endl;
    
    for(int i=0; i<extGridY.size(); ++i)
        o << extGridY[i] << " ";
    o << endl;

    
    o.close();
}

