// Copyright 2011-2012 Daniele Panozzo, panozzo@inf.ethz.ch. All rights reserved.
// The code in this file cannot be used for commercial
// applications without prior written permission from Daniele Panozzo

#ifndef LINUX
#include <QT/QImage.h>
#else
#include <QImage>
#endif
#include <Eigen/Core>
#include <QtOpenGL/QGLWidget>

// Enable the movie recording keys: o to start and p to stop
//#define MOVIEMODE

// Enable profiling
// #define TIMER

using namespace std;
using namespace Eigen;


class State
{
    public:

    static State* Instance();
    
    #define ENERGY_TYPE_NUM 3
    typedef enum {ENERGY_ASAP, ENERGY_ARAP, ENERGY_HYBRID} EnergyType;

    #define SALIENCY_TYPE_NUM 3
    typedef enum {SAL_USER, SAL_LOADED, SAL_BOTH} SaliencyType;

    QImage qImage;
    QImage tex;

    VectorXd gridX;      // current grid position
    VectorXd gridY;      // current grid position

    VectorXd gridXOri;      // original grid position
    VectorXd gridYOri;      // original grid position

    
    VectorXd gridU;      // fixed tex coordinates
    VectorXd gridV;      

    vector<unsigned char> saliency;   // saliency map
    vector<unsigned char> saliencyLoaded;   // saliency map
    vector<unsigned char> saliencyLoadedOriginal;

    VectorXi saliencyX;  // grid position for gradient cells
    VectorXi saliencyY;  // grid position for gradient cells
    MatrixXd integratedSaliency;
    MatrixXd integratedSaliencyValid;
    
    void loadImage(char* str);
    void loadSaliency(char* str);
    
    void computeGradient(const char* str);
    
    void init();
    void initGrid();
    void initSaliency(bool clear = true);
    void updateSaliency(); // send the saliency texture to the GPU and integrate the saliency
    void updateSaliencyLocal(); // send the saliency texture to the GPU and integrate the saliency only on the modified parts
    void updateLoaded(); // apply brightness and contrast to loaded saliency
    void updateCrop(); // called every time the crop changes
    void moveCrop(); // compute the crop based on the parameters in crop2P
    void updateSaliencyGrid(); // update the rectangular area where the saliency is computed
    void clearSaliency(); // clear the saliency

    void   setSal(int x, int y, unsigned char val);
    void   setSalLoaded(int x, int y, unsigned char val);
    void   setSalLoadedOriginal(int x, int y, unsigned char val);
    double getSal(int x, int y);
    double getSalLoaded(int x, int y);
    double getSalLoadedOriginal(int x, int y);
    
    Vector2d currentSize;    // current size of the image
    Vector2d originalSize;   // original size of the image

    Vector2d cropBL;   // bottom left corner of cropping rectangle
    Vector2d cropTR;   // top right corner of cropping rectangle

    Vector2d crop2P;   // simpler two parameters crop
    
    double ws;
    
    // brush options
    double brushIntensity;
    double brushRadius;
    bool brushSmooth;
    
    bool eraserActive;
    
    // textures
    GLuint textureImage;
    GLuint textureImageLinear;
    GLuint textureSaliency;
    GLuint textureSaliencyLoaded;
    
    // minimal edge lenght (%)
    double LWfactor;
    double LHfactor;
    
    void solve(bool force = false, bool render = true);
    void solveARAP();
    void solveASAP();
    void solveUniform();
    void generateARAPEnergySystem(MatrixXd& A, VectorXd& b, bool addLapTerms);
    void generateASAPEnergySystem(MatrixXd& A, VectorXd& b, bool addLapTerms);
    void generateHYBRIDEnergySystem(MatrixXd& A, VectorXd& b, double alpha, bool addLapTerms);
    void addLaplacianTerms(MatrixXd& A, VectorXd& b);
    
    EnergyType energyType;
    SaliencyType saliencyType;
    
    bool autoSolve;
    bool addLaplacianTermsEnabled; // Add laplacian terms to the energy
    double laplacianWeight; // Weight for the laplacian terms
    
    double HYBRID_alpha; // blend the two energies
    
    unsigned int XCELLS;
    unsigned int YCELLS;
    
    // generate and store the higher res grid used for rendering purposes
    double CubicInterpolate(double y0,double y1,double y2,double y3,double mu);
    double SplineCubicInterpolate(double y0,double y1,double y2,double y3,double mu);
    double LERP(double y0, double y1, double mu);
    void generateInterpolatedGrid(int n);
    void generateInterpolatedGrid_aux(VectorXd& gridX, VectorXd& gridU, VectorXd& extGridX, VectorXd& extGridU, double nsplit);

    VectorXd extGridX;
    VectorXd extGridY;
    VectorXd extGridU;
    VectorXd extGridV;      

    bool useExtGrid;
    int resExtGrid;
    
    string filename;
    
    bool exportMode;
    
    void saveSaliency(const char* str);
    void saveParams(const char* str);
    void saveMap(const char* str);
    
    double brightness;
    double contrast;
    
private:
    unsigned int roundPow2(unsigned int v);

    State();  // Private so that it can  not be called
    State(State const&){};             // copy constructor is private
    void operator=(State const&){};    // assignment operator is private
    static State* m_pInstance;

};
